;; Generated package description from satchel.el  -*- no-byte-compile: t -*-
(define-package "satchel" "0.2.0.20220223.202624" "A bag for your files, separated by git branches" '((emacs "27.2") (project "0.8.1")) :commit "d741c1f36cca51a64a2d02a58f9ae25f25f84e8b" :url "https://elpa.gnu.org/packages/satchel.html" :authors '(("Theodor Thornhill" . "theo@thornhill.no")) :maintainer '("Theodor Thornhill" . "theo@thornhill.no") :keywords '("tools" "git"))
